from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from sila2.framework import Command, Feature, FullyQualifiedIdentifier, Property
from sila2.server import FeatureImplementationBase

from .lockcontroller_types import LockServer_Responses, UnlockServer_Responses


class LockControllerBase(FeatureImplementationBase, ABC):

    """
    This Feature allows a SiLA Client to lock a SiLA Server for exclusive use, preventing other SiLA Clients
    from using the SiLA Server while it is locked. To lock a SiLA Server a Lock Identifier has to be set, using the
    'LockServer' command. This Lock Identifier has to be sent along with every (lock protected)
    request to the SiLA Server in order to use its functionality.

    To send the lock identifier the SiLA Client Meta Data 'LockIdentifier' has to be used.

    When locking a SiLA Server a timeout can be specified that defines the time after which the SiLA Server will
    be automatically unlocked if no request with a valid lock identifier has been received meanwhile.
    After the timeout has expired or after explicit unlock no lock identifier has to be sent any more.
    """

    @abstractmethod
    def get_IsLocked(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> bool:
        """
          Returns true if the SiLA Server is currently locked or false else.

        This property MUST NOT be lock protected, so that any SiLA Client can query the current lock state
        of a SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns true if the SiLA Server is currently locked or false else.

        This property MUST NOT be lock protected, so that any SiLA Client can query the current lock state
        of a SiLA Server.
        """
        pass

    @abstractmethod
    def LockServer(
        self, LockIdentifier: str, Timeout: int, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> LockServer_Responses:
        """
          Locks a SiLA Server for exclusive use by setting a lock identifier that has to be sent along with
        any following (lock protected) request as long as the SiLA Server is locked.
        The lock can be reset by issuing the 'Unlock Server' command.


          :param LockIdentifier: The lock identifier that has to be sent along with every (lock protected) request to use the server's
          functionality.

          :param Timeout: The time (in seconds) after a SiLA Server is automatically unlocked when no request with a valid lock identifier
          has been received meanwhile. A timeout of zero seconds specifies an infinite time (no timeout).

          :param metadata: The SiLA Client Metadata attached to the call

        """
        pass

    @abstractmethod
    def UnlockServer(
        self, LockIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> UnlockServer_Responses:
        """
          Unlocks a locked SiLA Server. No lock identifier has to be sent for any following calls until
        the server is locked again via the 'Lock Server' command.


          :param LockIdentifier: The lock identifier that has been used to lock the SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call

        """
        pass

    @abstractmethod
    def get_calls_affected_by_LockIdentifier(self) -> List[Union[Feature, Command, Property, FullyQualifiedIdentifier]]:
        """
          Returns the fully qualified identifiers of all features, commands and properties affected by the
          SiLA Client Metadata 'Delay'.

          **Description of 'LockIdentifier'**:
          The lock identifier has to be sent with every (lock protected) call in order to use the functionality
        of a locked SiLA Server.

          :return: Fully qualified identifiers of all features, commands and properties affected by the
              SiLA Client Metadata 'Delay'.
        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.framework.command.command import Command


class Response(NamedDataNode):
    fully_qualified_identifier: FullyQualifiedIdentifier
    parent_command: Command

    def __init__(self, fdl_node, parent_command: Command):
        super().__init__(
            fdl_node,
            parent_command.parent_feature,
            getattr(parent_command.parent_feature._pb2_module, f"{parent_command._identifier}_Responses"),
        )
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_command.fully_qualified_identifier}/Response/{self._identifier}"
        )



from typing import Any, Dict
from uuid import UUID

from sila2.features.authorizationproviderservice import (
    AuthorizationFailed,
    AuthorizationProviderServiceBase,
    InvalidAccessToken,
    Verify_Responses,
)
from sila2.framework import FullyQualifiedIdentifier
from sila2.server import SilaServer
from sila2.server.default_feature_implementations.authenticationservice_impl import AccessToken


class AuthorizationProviderServiceImpl(AuthorizationProviderServiceBase):
    tokens: Dict[UUID, AccessToken]

    def __init__(self, parent_server: SilaServer):
        self.server_uuid = parent_server.server_uuid
        self.tokens = parent_server.feature_servicers["AuthenticationService"].implementation.active_tokens

    def Verify(
        self,
        AccessToken: str,
        RequestedServer: str,
        RequestedFeature: str,
        *,
        metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> Verify_Responses:
        try:
            token = self.tokens[UUID(AccessToken)]
        except ValueError:
            raise InvalidAccessToken  # no UUID
        except KeyError:
            raise AuthorizationFailed  # unknown token

        if UUID(RequestedServer) != self.server_uuid:
            raise AuthorizationFailed

        if FullyQualifiedIdentifier(RequestedFeature) not in token.features:
            raise AuthorizationFailed

        if token.remaining_lifetime.total_seconds() <= 0:
            self.tokens.pop(token.token)
            raise AuthorizationFailed

        return Verify_Responses(int(token.remaining_lifetime.total_seconds()))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, List, Optional

from google.protobuf.message import Message

from sila2.framework.abc.constraint import Constraint
from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.parameter import Parameter
from sila2.framework.errors.validation_error import ValidationError
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class Constrained(DataType):
    base_type: DataType
    constraints: List[Constraint]

    def __init__(self, base_type: DataType, constraints: Iterable[Constraint]):
        self.base_type = base_type
        self.constraints = list(constraints)
        self.native_type = self.base_type.native_type
        self.message_type = self.base_type.message_type

    def to_native_type(self, message: Message, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Any:
        native_object = self.base_type.to_native_type(message, toplevel_named_data_node=toplevel_named_data_node)
        self.__validate(native_object, toplevel_named_data_node, can_raise_validation_error=True)
        return native_object

    def to_message(self, *args: Any, **kwargs: Any) -> Message:
        msg = self.base_type.to_message(*args, **kwargs)

        toplevel_named_data_node = kwargs.get("toplevel_named_data_node")
        native_object = self.base_type.to_native_type(msg, toplevel_named_data_node=toplevel_named_data_node)
        self.__validate(native_object, toplevel_named_data_node, can_raise_validation_error=False)

        return msg

    def __validate(
        self, value: Any, toplevel_named_data_node: Optional[NamedDataNode], can_raise_validation_error: bool
    ) -> None:
        for constraint in self.constraints:
            if not constraint.validate(value):
                if isinstance(toplevel_named_data_node, Parameter) and can_raise_validation_error:
                    raise ValidationError(toplevel_named_data_node, f"{constraint!r}: {value!r}")

    @staticmethod
    def from_fdl_node(fdl_node, parent_feature: Feature, parent_namespace) -> Constrained:
        dtype = DataType.from_fdl_node(xpath_sila(fdl_node, "sila:DataType")[0], parent_feature, parent_namespace)
        constraints = [
            Constraint.from_fdl_node(constraint_node, parent_feature, dtype)
            for constraint_node in xpath_sila(fdl_node, "sila:Constraints/*")
        ]
        return Constrained(dtype, constraints)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Integer as SilaInteger


class Integer(DataType):
    native_type = int
    message_type: Type[SilaInteger]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Integer

    def to_native_type(self, message: SilaInteger, toplevel_named_data_node: Optional[NamedDataNode] = None) -> int:
        return message.value

    def to_message(self, value: int, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaInteger:
        return self.message_type(value=value)

    @staticmethod
    def from_string(value: str) -> int:
        if not re.match("^[-+]?[0-9]+$", value):
            raise ValueError(f"Cannot parse as integer: '{value}'")
        return int(value)



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from typing import List as TypingList
from typing import Optional

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class List(DataType):
    element_type: DataType

    def __init__(self, fdl_node, parent_feature: Feature, parent_namespace):
        self.element_type = DataType.from_fdl_node(
            xpath_sila(fdl_node, "sila:DataType")[0], parent_feature, parent_namespace
        )
        self.native_type = TypingList[self.element_type.native_type]
        self.message_type = TypingList[self.element_type.message_type]  # repeated fields are no messages

    def to_native_type(
        self, message: RepeatedCompositeFieldContainer, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> Any:
        return [
            self.element_type.to_native_type(item, toplevel_named_data_node=toplevel_named_data_node)
            for item in message
        ]

    def to_message(self, items: TypingList, toplevel_named_data_node: Optional[NamedDataNode] = None) -> TypingList:
        return [self.element_type.to_message(item, toplevel_named_data_node=toplevel_named_data_node) for item in items]



from __future__ import annotations

from abc import ABC, abstractmethod
from base64 import standard_b64decode
from os.path import join
from typing import TYPE_CHECKING

import grpc

from sila2 import resource_dir
from sila2.framework.utils import run_protoc

pb2_module, _ = run_protoc(join(resource_dir, "proto", "SiLAFramework.proto"))

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError

    pb2_module = SiLAFramework_pb2


class SilaError(Exception, ABC):
    pb2_module: SiLAFramework_pb2 = pb2_module

    def __init__(self, message: str):
        super().__init__(message)

    @abstractmethod
    def to_message(self) -> SiLAError:
        pass

    @classmethod
    @abstractmethod
    def from_message(cls, message: SiLAError, client: SilaClient) -> SiLAError:
        pass

    @staticmethod
    def is_sila_error(exception: Exception) -> bool:
        if not isinstance(exception, grpc.RpcError):
            return False

        if not exception.code() == grpc.StatusCode.ABORTED:
            return False

        try:
            SilaError.pb2_module.SiLAError.FromString(standard_b64decode(exception.details()))
        except:  # noqa: 722
            return False

        return True

    @staticmethod
    def from_rpc_error(rpc_error: grpc.RpcError, client: SilaClient):
        if not SilaError.is_sila_error(rpc_error):
            raise ValueError("Error is no SiLAError")

        sila_err = SilaError.pb2_module.SiLAError.FromString(standard_b64decode(rpc_error.details()))

        if sila_err.HasField("validationError"):
            from sila2.framework.errors.validation_error import ValidationError

            return ValidationError.from_message(sila_err.validationError, client)
        if sila_err.HasField("definedExecutionError"):
            from sila2.framework.errors.defined_execution_error import DefinedExecutionError

            return DefinedExecutionError.from_message(sila_err.definedExecutionError, client)
        if sila_err.HasField("undefinedExecutionError"):
            from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError

            return UndefinedExecutionError.from_message(sila_err.undefinedExecutionError, client)
        if sila_err.HasField("frameworkError"):
            from sila2.framework.errors.framework_error import FrameworkError

            return FrameworkError.from_message(sila_err.frameworkError, client)

        raise NotImplementedError(f"SiLAError type not supported: {sila_err}")  # should not happen



from typing import TYPE_CHECKING

from .silaservice_base import SiLAServiceBase
from .silaservice_errors import UnimplementedFeature
from .silaservice_feature import SiLAServiceFeature
from .silaservice_types import GetFeatureDefinition_Responses, SetServerName_Responses

__all__ = [
    "SiLAServiceBase",
    "SiLAServiceFeature",
    "GetFeatureDefinition_Responses",
    "SetServerName_Responses",
    "UnimplementedFeature",
]

if TYPE_CHECKING:
    from .silaservice_client import SiLAServiceClient  # noqa: F401

    __all__.append("SiLAServiceClient")



from __future__ import annotations

from typing import NamedTuple


class GetFeatureDefinition_Responses(NamedTuple):

    FeatureDefinition: str
    """
    The Feature definition in XML format (according to the Feature Definition Schema).
    """


class SetServerName_Responses(NamedTuple):

    pass



import os
import sys
from os.path import dirname, join

if sys.version_info.minor == 7:
    import importlib_metadata as metadata
else:
    from importlib import metadata

__version__ = metadata.version(__name__)


# this line has to be executed before protobuf is imported
# https://github.com/protocolbuffers/protobuf/issues/3002#issuecomment-325459597
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

resource_dir = join(dirname(__file__), "resources")


__all__ = ["__version__", "resource_dir"]



from sila2.server.feature_implementation_base import FeatureImplementationBase
from sila2.server.metadata_interceptor import MetadataInterceptor
from sila2.server.observables.observable_command_instance import (
    ObservableCommandInstance,
    ObservableCommandInstanceWithIntermediateResponses,
)
from sila2.server.sila_server import SilaServer

__all__ = [
    "FeatureImplementationBase",
    "MetadataInterceptor",
    "ObservableCommandInstance",
    "ObservableCommandInstanceWithIntermediateResponses",
    "SilaServer",
]



import warnings
from typing import Dict
from uuid import UUID

from zeroconf import ServiceBrowser, ServiceInfo, ServiceListener, Zeroconf

from sila2.client.sila_client import SilaClient


class SilaServiceListener(ServiceListener):
    parent_browser: ServiceBrowser
    services: Dict[str, SilaClient]

    def __init__(self, parent_browser: ServiceBrowser):
        self.parent_browser = parent_browser
        self.services = {}

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        self.__add_client(info)

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        self.__add_client(info)

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        if name in self.services:
            self.services.pop(name)

    def __add_client(self, service_info: ServiceInfo) -> None:
        service_name = service_info.name
        ip_address = service_info.parsed_addresses()[0]
        try:
            self.services[service_name] = SilaClient(ip_address, service_info.port)
        except Exception as ex:
            service_uuid = UUID(service_name.split(".")[0])
            warnings.warn(
                RuntimeWarning(
                    f"SiLA Server Discovery found a service with UUID {service_uuid} but failed to connect to it: "
                    f"{ex.__class__.__name__} - {ex}"
                )
            )



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable

from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier


class MetadataInterceptor(ABC):
    def __init__(self, required_metadata: Iterable[FullyQualifiedIdentifier]) -> None:
        self.required_metadata = frozenset(required_metadata)

    @abstractmethod
    def intercept(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target_call: FullyQualifiedIdentifier
    ) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type
from uuid import UUID

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.intermediate_response import IntermediateResponse
from sila2.framework.command.parameter import Parameter
from sila2.framework.command.response import Response
from sila2.framework.property.property import Property

if TYPE_CHECKING:
    from sila2.framework.feature import Feature
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Binary as SilaBinary


class Binary(DataType):
    native_type = bytes
    parent_feature: Feature
    binary_transfer_possible: bool
    MAX_SIZE = 2 * 1024 ** 2
    message_type: Type[SilaBinary]

    def __init__(self, parent_feature: Feature, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Binary
        self.parent_feature = parent_feature

    def to_native_type(self, message: SilaBinary, toplevel_named_data_node: Optional[NamedDataNode] = None) -> bytes:
        if message.HasField("value"):
            return message.value

        if not message.HasField("binaryTransferUUID"):
            raise ValueError("Binary message has neigher value, not binaryTransferUUID")

        if self.parent_feature._binary_transfer_handler is None:
            raise NotImplementedError("Feature reference does not implement Binary Transfer")

        binary_uuid = UUID(message.binaryTransferUUID)
        if isinstance(toplevel_named_data_node, (Parameter, Property, Response, IntermediateResponse)):
            return self.parent_feature._binary_transfer_handler.to_native_type(binary_uuid, toplevel_named_data_node)

        raise ValueError("Binary Transfer only applies to Command Parameters, Responses and Intermediate Responses")

    def to_message(self, value: bytes, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaBinary:
        if not isinstance(value, bytes):
            raise TypeError("Expected a bytes value")

        if len(value) <= self.MAX_SIZE:
            return self.message_type(value=value)

        if isinstance(toplevel_named_data_node, (Parameter, Property, Response, IntermediateResponse)):
            if self.parent_feature._binary_transfer_handler is None:
                raise NotImplementedError("Feature reference does not implement Binary Transfer")
            return self.parent_feature._binary_transfer_handler.to_message(
                value, self.message_type, toplevel_named_data_node
            )

        raise ValueError("Binary Transfer only applies to Command Parameters, Responses and Intermediate Responses")



from sila2.framework.command.command import Command
from sila2.framework.command.execution_info import CommandExecutionInfo, CommandExecutionStatus
from sila2.framework.errors.command_execution_not_accepted import CommandExecutionNotAccepted
from sila2.framework.errors.command_execution_not_finished import CommandExecutionNotFinished
from sila2.framework.errors.defined_execution_error import DefinedExecutionError
from sila2.framework.errors.invalid_command_execution_uuid import InvalidCommandExecutionUUID
from sila2.framework.errors.invalid_metadata import InvalidMetadata
from sila2.framework.errors.no_metadata_allowed import NoMetadataAllowed
from sila2.framework.errors.sila_connection_error import SilaConnectionError
from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError
from sila2.framework.errors.validation_error import ValidationError
from sila2.framework.feature import Feature
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.metadata import Metadata
from sila2.framework.property.property import Property

__all__ = [
    "Command",
    "CommandExecutionInfo",
    "CommandExecutionStatus",
    "CommandExecutionNotAccepted",
    "CommandExecutionNotFinished",
    "DefinedExecutionError",
    "InvalidCommandExecutionUUID",
    "InvalidMetadata",
    "NoMetadataAllowed",
    "SilaConnectionError",
    "UndefinedExecutionError",
    "ValidationError",
    "Feature",
    "FullyQualifiedIdentifier",
    "Metadata",
    "Property",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Union

from sila2.framework import Command, Feature, FullyQualifiedIdentifier, Property
from sila2.server import FeatureImplementationBase


class AuthorizationServiceBase(FeatureImplementationBase, ABC):

    """
    This Feature provides access control for the implementing server.

    It specifies the SiLA Client Metadata for the access token, that has been provided by the
    AuthenticationService core Feature.
    """

    @abstractmethod
    def get_calls_affected_by_AccessToken(self) -> List[Union[Feature, Command, Property, FullyQualifiedIdentifier]]:
        """
        Returns the fully qualified identifiers of all features, commands and properties affected by the
        SiLA Client Metadata 'Delay'.

        **Description of 'AccessToken'**:
        Token to be sent with every call in order to get access to the SiLA Server functionality.

        :return: Fully qualified identifiers of all features, commands and properties affected by the
            SiLA Client Metadata 'Delay'.
        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.command.parameter import Parameter
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError
    from sila2.pb2_stubs.SiLAFramework_pb2 import ValidationError as SilaValidationError


class ValidationError(SilaError):
    parameterfully_qualified_identifier: FullyQualifiedIdentifier
    message: str

    def __init__(self, parameter: Parameter, message: str):
        self.parameterfully_qualified_identifier = parameter.fully_qualified_identifier
        self.message = message
        super().__init__(
            f"Constraint validation failed for parameter {self.parameterfully_qualified_identifier}: {message}"
        )

    def to_message(self) -> SiLAError:
        return self.pb2_module.SiLAError(
            validationError=self.pb2_module.ValidationError(
                parameter=self.parameterfully_qualified_identifier, message=self.message
            )
        )

    @classmethod
    def from_message(cls, message: SilaValidationError, client: SilaClient) -> ValidationError:
        return cls(
            client._children_by_fully_qualified_identifier[FullyQualifiedIdentifier(message.parameter)], message.message
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from zeroconf import Zeroconf

from sila2.discovery.service_info import SilaServiceInfo

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer


class SilaServiceBroadcaster:
    zc: Zeroconf

    def __init__(self):
        self.zc = Zeroconf()

    def register_server(self, server: SilaServer, address: str, port: int) -> SilaServiceInfo:
        service_info = SilaServiceInfo(server, address, port)
        self.zc.register_service(service_info)
        return service_info

    def unregister_server(self, info: SilaServiceInfo) -> None:
        self.zc.unregister_service(info)



from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Union

from sila2.framework.constraints.comparison_constraint import ComparisonConstraint
from sila2.framework.data_types.date import Date
from sila2.framework.data_types.integer import Integer
from sila2.framework.data_types.real import Real
from sila2.framework.data_types.string import String
from sila2.framework.data_types.time import Time
from sila2.framework.data_types.timestamp import Timestamp

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class MaximalExclusive(ComparisonConstraint):
    def __init__(self, value: str, base_type: Union[String, Integer, Real, Date, Time, Timestamp]):
        super().__init__(base_type, operator.gt, value)

    @classmethod
    def from_fdl_node(
        cls, fdl_node, parent_feature: Feature, base_type: Union[String, Integer, Real, Date, Time, Timestamp]
    ) -> MaximalExclusive:
        return cls(fdl_node.text, base_type)



from typing import Callable, Iterator, TypeVar

import grpc
from grpc import RpcError
from grpc._channel import _MultiThreadedRendezvous

T = TypeVar("T")


class SubscriptionStream(Iterator[T]):
    __wrapped_stream: _MultiThreadedRendezvous
    __converter_func: Callable

    def __init__(self, wrapped_stream: _MultiThreadedRendezvous, converter_func: Callable) -> None:
        self.__wrapped_stream = wrapped_stream
        self.__converter_func = converter_func

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.__converter_func(next(self.__wrapped_stream))
        except RpcError as rpc_err:
            if rpc_err.code() == grpc.StatusCode.CANCELLED:
                raise StopIteration()
            raise

    def cancel(self):
        self.__wrapped_stream.cancel()



from typing import TYPE_CHECKING

from .lockcontroller_base import LockControllerBase
from .lockcontroller_errors import InvalidLockIdentifier, ServerAlreadyLocked, ServerNotLocked
from .lockcontroller_feature import LockControllerFeature
from .lockcontroller_types import LockServer_Responses, UnlockServer_Responses

__all__ = [
    "LockControllerBase",
    "LockControllerFeature",
    "LockServer_Responses",
    "UnlockServer_Responses",
    "InvalidLockIdentifier",
    "ServerAlreadyLocked",
    "ServerNotLocked",
]

if TYPE_CHECKING:
    from .lockcontroller_client import LockControllerClient  # noqa: F401

    __all__.append("LockControllerClient")



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Duration as SilaDuration


class Duration(MessageMappable):
    native_type = timedelta
    message_type: Type[SilaDuration]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Duration

    def to_native_type(
        self, message: SilaDuration, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> timedelta:
        return timedelta(seconds=message.seconds, microseconds=round(message.nanos / 1000))

    def to_message(self, duration: timedelta, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaDuration:
        seconds, rest = divmod(duration.total_seconds(), 1)
        return self.message_type(seconds=round(seconds), nanos=round(rest * 1e9))



from __future__ import annotations

from typing import TYPE_CHECKING

from zeroconf import ServiceInfo

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer


class SilaServiceInfo(ServiceInfo):
    sila_server: SilaServer

    def __init__(self, server: SilaServer, address: str, port: int):
        super().__init__(
            type_="_sila._tcp.local.",
            name=f"{server.server_uuid}._sila._tcp.local.",
            parsed_addresses=[address],
            port=port,
            properties=dict(
                version=server.server_version,
                server_name=server.server_name,
                description=server.server_description,
            ),
        )
        self.sila_server = server



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.data_types.timezone import Timezone

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Timestamp as SilaTimestamp


class Timestamp(DataType):
    native_type = datetime
    message_type: Type[SilaTimestamp]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Timestamp
        self.__timezone_field = Timezone(silaframework_pb2_module)

    def to_message(self, dt: datetime, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaTimestamp:
        if not isinstance(dt, datetime):
            raise TypeError("Expected a datetime")
        if dt.tzinfo is None:
            raise ValueError("No timezone provided")

        return self.message_type(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            timezone=self.__timezone_field.to_message(dt.tzinfo),
        )

    def to_native_type(
        self, message: SilaTimestamp, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> datetime:
        return datetime(
            year=message.year,
            month=message.month,
            day=message.day,
            hour=message.hour,
            minute=message.minute,
            second=message.second,
            tzinfo=self.__timezone_field.to_native_type(message.timezone),
        )

    @staticmethod
    def from_string(value: str) -> datetime:
        if len(value) not in (20, 25):
            raise ValueError(
                f"Invalid timestamp format: '{value}'. "
                f"Must be 'YYYY-MM-DDTHH:MM:SS' plus timezone as 'Z', '-HH:MM' or '+HH:MM'"
            )
        dt = datetime.fromisoformat(value[:19])
        tz = Timezone.from_string(value[19:])
        return datetime(
            year=dt.year, month=dt.month, day=dt.day, hour=dt.hour, minute=dt.minute, second=dt.second, tzinfo=tz
        )



import grpc
from grpc import RpcError


class SilaConnectionError(Exception):
    inner_exception: Exception

    def __init__(self, inner_exception: Exception):
        self.inner_exception = inner_exception
        if isinstance(inner_exception, RpcError):
            if inner_exception.code() == grpc.StatusCode.UNAVAILABLE:
                message = "Failed to establish connection to the server"
            else:
                message = f"{inner_exception.code().name} - {inner_exception.details()}"
            message = f"{inner_exception.code().name} - {message}"
        else:
            message = str(inner_exception)
        super().__init__(f"{inner_exception.__class__.__name__}: {message}")



from __future__ import annotations

from abc import ABC
from datetime import date, datetime, time, timezone
from typing import Any, Callable, Tuple, Union

from sila2.framework.abc.constraint import Constraint
from sila2.framework.data_types.date import Date
from sila2.framework.data_types.integer import Integer
from sila2.framework.data_types.real import Real
from sila2.framework.data_types.string import String
from sila2.framework.data_types.time import Time
from sila2.framework.data_types.timestamp import Timestamp


class ComparisonConstraint(Constraint, ABC):
    comparison_operator: Callable[[Any, Any], bool]
    base_type: Union[String, Integer, Real, Date, Time, Timestamp]
    parsed_value: Union[str, int, float, datetime]
    value: str

    def __init__(
        self,
        base_type: Union[String, Integer, Real, Date, Time, Timestamp],
        comparison_operator: Callable[[Any, Any], bool],
        value: str,
    ):
        self.base_type = base_type
        self.comparison_operator = comparison_operator
        self.value = value
        self.parsed_value = self._parse_value_from_string(value)

    def validate(self, value: Union[str, int, float, Tuple[date, timezone], time, datetime]):
        parsed_value = self._convert_value_for_comparison(value)
        return self.comparison_operator(self.parsed_value, parsed_value)

    def _convert_value_for_comparison(self, value: Union[str, int, float, Tuple[date, timezone], time, datetime]):
        if isinstance(self.base_type, (String, Integer, Real, Timestamp)):
            return value
        if isinstance(self.base_type, Date):
            d, tz = value
            return datetime(year=d.year, month=d.month, day=d.day, tzinfo=tz)
        if isinstance(self.base_type, Time):
            return datetime(
                year=2000,
                month=1,
                day=1,
                hour=value.hour,
                minute=value.minute,
                second=value.second,
                tzinfo=value.tzinfo,
            )
        else:
            raise NotImplementedError  # should never happen

    def _parse_value_from_string(self, value: str) -> Union[str, int, float, datetime]:
        if isinstance(self.base_type, String):
            return value
        if isinstance(self.base_type, Integer):
            from sila2.framework.constraints.maximal_exclusive import MaximalExclusive
            from sila2.framework.constraints.maximal_inclusive import MaximalInclusive
            from sila2.framework.constraints.minimal_exclusive import MinimalExclusive
            from sila2.framework.constraints.minimal_inclusive import MinimalInclusive

            if isinstance(self, (MaximalExclusive, MaximalInclusive, MinimalExclusive, MinimalInclusive)):
                return float(value)
            return int(value)
        if isinstance(self.base_type, Real):
            return float(value)
        if isinstance(self.base_type, Date):
            d, tz = Date.from_string(value)
            return datetime(year=d.year, month=d.month, day=d.day, tzinfo=tz)
        if isinstance(self.base_type, Time):
            t = Time.from_string(value)
            return datetime(year=2000, month=1, day=1, hour=t.hour, minute=t.minute, second=t.second, tzinfo=t.tzinfo)
        if isinstance(self.base_type, Timestamp):
            return Timestamp.from_string(value)
        else:
            raise NotImplementedError  # should never happen

    def __repr__(self) -> str:
        if isinstance(self.base_type, (Integer, Real)):
            return f"{self.__class__.__name__}({self.parsed_value})"
        return f"{self.__class__.__name__}({self.value!r})"



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import DefinedExecutionError as SilaDefinedExecutionError
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError


class DefinedExecutionError(SilaError):
    message: str
    identifier: str
    fully_qualified_identifier: FullyQualifiedIdentifier

    def __init__(self, error_node: DefinedExecutionErrorNode, message: str):
        self.message = message
        self.identifier = error_node._identifier
        self.fully_qualified_identifier = error_node.fully_qualified_identifier

        # if self is a specific subclass
        if self.__class__.__name__ == error_node._identifier:
            super().__init__(message)
        else:
            super().__init__(f"{error_node._identifier}: {message}")

    def to_message(self) -> SiLAError:
        return self.pb2_module.SiLAError(
            definedExecutionError=self.pb2_module.DefinedExecutionError(
                errorIdentifier=self.fully_qualified_identifier, message=self.message
            )
        )

    @classmethod
    def from_message(cls, message: SilaDefinedExecutionError, client: SilaClient) -> DefinedExecutionError:
        return cls(
            client._children_by_fully_qualified_identifier[FullyQualifiedIdentifier(message.errorIdentifier)],
            message.message,
        )



import re
from distutils.dir_util import copy_tree
from glob import glob
from os import makedirs
from os.path import basename, dirname, join
from shutil import copy

from setuptools import find_packages, setup

base_dir = dirname(__file__)
package_dir = "src"
name = "sila2"

test_requirements = [
    "pytest",
    "pytest-cov",  # pytest: code coverage
]
code_quality_requirements = [
    "flake8",  # style checker
]


def copy_resources_from_sila_base():
    resource_dir = join(base_dir, package_dir, name, "resources")
    makedirs(resource_dir, exist_ok=True)

    # xsd
    copy_tree(join(base_dir, "sila_base", "schema"), join(resource_dir, "xsd"))

    # xsl
    makedirs(join(resource_dir, "xsl"), exist_ok=True)
    for file in glob(join(base_dir, "sila_base", "xslt", "*.xsl")):
        copy(file, join(resource_dir, "xsl", basename(file)))

    # proto
    makedirs(join(resource_dir, "proto"), exist_ok=True)
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLAFramework.proto"),
        join(resource_dir, "proto", "SiLAFramework.proto"),
    )
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLABinaryTransfer.proto"),
        join(resource_dir, "proto", "SiLABinaryTransfer.proto"),
    )


def prepare_readme():
    """README contains gitlab-internal links, which need to be extended for PyPI"""
    content = open(join(dirname(__file__), "README.md")).read()
    link_pattern = r"\]\(([^h][^)]+)\)"
    return re.sub(link_pattern, "](https://gitlab.com/sila2/sila_python/-/tree/master/\\1)", content)


copy_resources_from_sila_base()

setup(
    name=name,
    version="0.3.5",
    author="Niklas Mertsch",
    author_email="niklas.mertsch@stud.uni-goettingen.de",
    description="Python implementation of the SiLA 2 standard for lab automation",
    long_description=prepare_readme(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/sila2/sila_python",
    project_urls={
        "Source": "https://gitlab.com/sila2/sila_python",
        "Bug Reports": "https://gitlab.com/sila2/sila_python/-/issues",
        "SiLA Standard": "https://sila-standard.org",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(where=package_dir),
    package_dir={"": package_dir},
    install_requires=[
        "black",
        "grpcio",
        "grpcio-tools",
        "isort",
        "jinja2",
        "lxml",
        "jsonschema",
        "xmlschema",
        "zeroconf",
        "importlib-metadata; python_version < '3.8'",
    ],
    python_requires=">=3.7",
    extras_require=dict(
        tests=test_requirements,
        dev=test_requirements + code_quality_requirements,
    ),
    include_package_data=True,
)



from __future__ import annotations

from sila2.framework.abc.composite_message_mappable import CompositeMessageMappable
from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.utils import xpath_sila


class StructureElement(NamedDataNode):
    data_type: DataType

    def __init__(self, fdl_node, parent_feature, parent_namespace):
        super().__init__(fdl_node, parent_feature, parent_namespace)


class Structure(CompositeMessageMappable[StructureElement], DataType):
    def __init__(self, fdl_node, parent_feature, parent_namespace):
        message_type = getattr(
            parent_namespace,
            xpath_sila(fdl_node, "ancestor::*/sila:Identifier/text()")[-1] + "_Struct",
        )
        super().__init__(
            [StructureElement(node, parent_feature, message_type) for node in xpath_sila(fdl_node, "sila:Element")],
            getattr(parent_namespace, xpath_sila(fdl_node, "ancestor::*/sila:Identifier/text()")[-1] + "_Struct"),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Type

from google.protobuf.message import Message

if TYPE_CHECKING:
    from sila2.framework.abc.named_data_node import NamedDataNode


class MessageMappable(ABC):
    """Abstract class all classes representing Python-mappings for protobuf messages"""

    message_type: Type[Message]
    native_type: Type

    @abstractmethod
    def to_native_type(self, message: Message, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Any:
        """Convert a protobuf message to the associated native type"""
        pass

    @abstractmethod
    def to_message(self, *args, **kwargs) -> Message:
        """
        Construct a protobuf message from native objects

        Special kwargs:
        - toplevel_named_data_node: Optional[NamedDataNode]: The NamedDataNode object that the generated message is
            a part of (e.g. a Parameter instance)
        """
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from sila2.framework import FullyQualifiedIdentifier
from sila2.server import FeatureImplementationBase

from .authenticationservice_types import Login_Responses, Logout_Responses


class AuthenticationServiceBase(FeatureImplementationBase, ABC):

    """
    This Feature provides SiLA Clients with access tokens based on a user identification and password.
    1. the user needs to login with the Login command into the server with a user identification (=user name) and a password
    2. after verification, an Access Token with the Token Lifetime information will be generated and provided by the server.
    3. the user can log-out from the server with the Logout command - a valid Access Token is required to run this command.
    """

    @abstractmethod
    def Login(
        self,
        UserIdentification: str,
        Password: str,
        RequestedServer: str,
        RequestedFeatures: List[str],
        *,
        metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> Login_Responses:
        """
        Provides an access token based on user information.


        :param UserIdentification: The user identification string (e.g. a user name)

        :param Password: The password

        :param RequestedServer: The ServerUUID of the server for which an authorization is requested.

        :param RequestedFeatures: The fully qualified identifiers of features that are requested to access. If no feature is provided, this means that all features are requested.

        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - AccessToken: The token to be used along with accessing a Command or Property on a SiLA Server.

            - TokenLifetime: The lifetime (in seconds) of the provided access token as the maximum validity period after the last SiLA Server
        request.


        """
        pass

    @abstractmethod
    def Logout(self, AccessToken: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> Logout_Responses:
        """
        Invalidates the given access token immediately.


        :param AccessToken: The access token to be invalidated.

        :param metadata: The SiLA Client Metadata attached to the call

        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Type

import grpc

from sila2.client.client_feature import ClientFeature
from sila2.framework import DefinedExecutionError
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.features.silaservice import SiLAServiceClient
    from sila2.framework.utils import HasFullyQualifiedIdentifier


class SilaClient:
    SiLAService: SiLAServiceClient

    _channel: grpc.Channel
    _features: Dict[str, ClientFeature]
    _children_by_fully_qualified_identifier: Dict[FullyQualifiedIdentifier, HasFullyQualifiedIdentifier]
    _registered_defined_execution_error_classes: Dict[FullyQualifiedIdentifier, Type[DefinedExecutionError]]

    def __init__(self, hostname: str, port: int):
        self._channel = grpc.insecure_channel(f"{hostname}:{port}")
        self._features = {}
        self._children_by_fully_qualified_identifier = {}
        self._registered_defined_execution_error_classes = {}

        # import locally to prevent circular import
        from sila2.features.silaservice import SiLAServiceFeature, UnimplementedFeature

        self.__add_feature(SiLAServiceFeature._feature_definition)
        self._register_defined_execution_error_class(
            SiLAServiceFeature.defined_execution_errors["UnimplementedFeature"], UnimplementedFeature
        )

        for feature_id in self.SiLAService.ImplementedFeatures.get():
            if feature_id == self._features["SiLAService"].fully_qualified_identifier:
                continue
            self.__add_feature(self.SiLAService.GetFeatureDefinition(feature_id).FeatureDefinition)

    def __add_feature(self, feature_definition: str) -> None:
        feature = ClientFeature(feature_definition, self)
        self._children_by_fully_qualified_identifier[feature.fully_qualified_identifier] = feature
        self._children_by_fully_qualified_identifier.update(feature._children_by_fully_qualified_identifier)
        self._features[feature._identifier] = feature

        setattr(self, feature._identifier, feature)

    def _register_defined_execution_error_class(
        self, error_node: DefinedExecutionErrorNode, error_class: Type[DefinedExecutionError]
    ):
        self._registered_defined_execution_error_classes[error_node.fully_qualified_identifier] = error_class



import platform
import sys
from argparse import ArgumentParser, Namespace
from glob import glob
from os.path import dirname, exists, isdir, join
from typing import List, Optional

import sila2.code_generator
from sila2.code_generator.code_generator import CodeGenerator
from sila2.code_generator.feature_generator import FeatureGenerator
from sila2.framework import Feature


def _get_feature_definition_paths(feature_definition_paths: List[str]) -> List[str]:
    if platform.system() != "Windows":
        return feature_definition_paths

    paths = []
    for path in feature_definition_paths:
        files = glob(path, recursive=True)
        if not files:
            raise FileNotFoundError(f"No such files: '{path}'")
        paths.extend(files)
    return paths


def package(args: Namespace) -> None:
    package_name: str = args.package_name
    out_dir: str = args.output_dir
    overwrite: bool = args.overwrite
    feature_definitions: List[str] = _get_feature_definition_paths(args.feature_definitions)
    lock_controller: bool = args.lock_controller
    auth_features: bool = args.auth_features
    for_git_repo: bool = args.git

    features = [Feature(open(fdl_file).read()) for fdl_file in sorted(feature_definitions)]
    generator = CodeGenerator(overwrite=overwrite)
    generator.generate_package(
        package_name,
        features,
        out_dir,
        lock_controller=lock_controller,
        auth_features=auth_features,
        for_git_repo=for_git_repo,
    )
    generator.format_generated_files()


def add_features(args: Namespace) -> None:
    package_dir: str = args.package_dir
    new_fdl_files: List[str] = _get_feature_definition_paths(args.feature_definitions)

    if not exists(join(package_dir, "setup.py")):
        print("Given directory is not a package (does not contain setup.py)", file=sys.stderr)
        return

    generated_dir = find_matching_directory(join(package_dir, "*", "generated"))
    implementation_dir = find_matching_directory(join(package_dir, "*", "feature_implementations"))

    existing_fdl_files = glob(join(generated_dir, "*", "*.sila.xml"))

    old_features = {}
    for fdl_file in existing_fdl_files:
        feature = Feature(open(fdl_file).read())
        old_features[feature._identifier] = feature

    new_features = {}
    for fdl_file in new_fdl_files:
        feature = Feature(open(fdl_file).read())
        if feature._identifier in old_features:
            raise ValueError(f"Feature {feature._identifier} already exists. Use 'update' instead.")
        new_features[feature._identifier] = feature

    features = list(old_features.values()) + list(new_features.values())
    features.sort(key=lambda f: f._identifier)

    generator = CodeGenerator(overwrite=True)
    generator.generate_generated_dir(features, generated_dir)
    generator.generate_implementations(list(new_features.values()), implementation_dir)
    generator.format_generated_files()


def update(args: Namespace) -> None:
    package_dir: str = args.package_dir

    generated_dir = find_matching_directory(join(package_dir, "*", "generated"))
    implementation_dir = find_matching_directory(join(package_dir, "*", "feature_implementations"))

    features = []
    generator = CodeGenerator(overwrite=True)
    for fdl_file in sorted(glob(join(generated_dir, "*", "*.sila.xml"))):
        feature = Feature(open(fdl_file).read())
        feature_generator = FeatureGenerator(feature, overwrite=True)
        features.append(feature)

        feature_dir = dirname(fdl_file)
        old_base = open(f"{feature_dir}/{feature._identifier.lower()}_base.py").read()

        feature_generator.generate_feature_files(feature_dir)
        new_base = open(f"{feature_dir}/{feature._identifier.lower()}_base.py").read()

        # if base class changed: generate new impl
        if old_base != new_base:
            feature_generator.generate_impl(implementation_dir, prefix="updated_")
        generator.generated_files.extend(feature_generator.generated_files)
    generator._generate_client(features, generated_dir)
    generator.format_generated_files()


def find_matching_directory(pattern: str) -> str:
    matches = glob(pattern)
    if not matches:
        raise ValueError(f"No matching directory found: {pattern}")
    if len(matches) > 1:
        raise ValueError(f"Multiple matches found: {pattern}")

    match = matches[0]
    if not isdir(match):
        raise NotADirectoryError(f"Not a directory: {match}")
    return match


def main(argv: Optional[List[str]] = None) -> int:
    parser = ArgumentParser(
        prog=sila2.code_generator.__name__,
        description="SiLA 2 Python Code Generator",
    )
    parser.add_argument("--debug", action="store_true", help="Display more detailed error messages")
    commands = parser.add_subparsers(title="Commands", required=True, dest="command")

    # package generation subparser
    package_parser = commands.add_parser(
        "new-package", description="Generate a SiLA 2 Server/Client Python package from given feature definitions"
    )
    package_parser.add_argument("-n", "--package-name", help="Package name")
    package_parser.add_argument("-o", "--output-dir", default=".", help="Package directory (will contain 'setup.py')")
    package_parser.add_argument("--overwrite", action="store_true", help="Overwrite existing files")
    package_parser.add_argument("--git", action="store_true", help="Generate .gitignore and README.md")
    package_parser.add_argument(
        "--lock-controller", action="store_true", help="Add default implementation of the core feature LockController"
    )
    package_parser.add_argument(
        "--auth-features",
        action="store_true",
        help="Add default implementations of the core features "
        "AuthenticationService, AuthorizationProviderService and AuthorizationService",
    )
    package_parser.add_argument("feature_definitions", nargs="*", help="SiLA 2 feature definitions")
    package_parser.set_defaults(func=package)

    # feature addition subparser
    add_features_parser = commands.add_parser(
        "add-features", description="Add features to previously generated package"
    )
    add_features_parser.add_argument("-d", "--package-dir", default=".", help="Package directory (contains 'setup.py')")
    add_features_parser.add_argument("feature_definitions", nargs="*", help="SiLA 2 feature definitions")
    add_features_parser.set_defaults(func=add_features)

    # package update subparser
    update_parser = commands.add_parser(
        "update",
        description="Update a previously generated package after modifications to the feature definitions "
        "(refreshes the 'generated' submodule)",
    )
    update_parser.add_argument("-d", "--package-dir", default=".", help="Package directory (contains 'setup.py')")
    update_parser.set_defaults(func=update)

    args = parser.parse_args(argv)

    try:
        args.func(args)
        return 0
    except Exception as ex:
        if args.debug:
            raise ex

        print(f"{ex.__class__.__name__}: {ex}", file=sys.stderr)
        return 1



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Tuple, Union

import grpc
from google.protobuf.message import Message

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.command.command import Command
from sila2.framework.errors.defined_execution_error import DefinedExecutionError
from sila2.framework.errors.sila_connection_error import SilaConnectionError
from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.property.property import Property

if TYPE_CHECKING:
    from sila2.client.client_metadata import ClientMetadataInstance
    from sila2.client.sila_client import SilaClient


def prepare_metadata_for_grpc(metadata: Optional[Iterable[ClientMetadataInstance]]) -> Tuple[Tuple[str, bytes], ...]:
    # grpc metadata must be a n-tuple of 2-tuples (key/value pairs)
    if metadata is None:
        return tuple()

    metadata_tuples = []
    for meta_instance in metadata:
        metadata_tuples.append(
            (meta_instance.metadata.to_grpc_header_key(), meta_instance.metadata.to_message(meta_instance.value))
        )
    print(metadata_tuples)
    return tuple(metadata_tuples)


def call_rpc_function(
    rpc_func: Any,
    parameter_message: Message,
    metadata: Optional[Iterable[ClientMetadataInstance]],
    client: SilaClient,
    origin: Optional[Union[Command, Property]],
):
    # collect allowed defined execution errors
    allowed_errors: List[FullyQualifiedIdentifier] = []
    if origin is not None:
        allowed_errors.extend(e.fully_qualified_identifier for e in origin.defined_execution_errors)
    if metadata is not None:
        for meta_instance in metadata:
            allowed_errors.extend(e.fully_qualified_identifier for e in meta_instance.metadata.defined_execution_errors)

    # call rpc function
    try:
        if hasattr(rpc_func, "with_call"):  # don't know when this is the case, but one of the two always seems to work
            response_msg, _ = rpc_func.with_call(parameter_message, metadata=prepare_metadata_for_grpc(metadata))
        else:
            response_msg = rpc_func(parameter_message, metadata=prepare_metadata_for_grpc(metadata))
        return response_msg
    except Exception as ex:
        # raise appropriate exception
        if isinstance(ex, grpc.RpcError) and ex.code() == grpc.StatusCode.UNIMPLEMENTED:
            raise NotImplementedError(ex.details())

        if SilaError.is_sila_error(ex):
            sila_err = SilaError.from_rpc_error(ex, client=client)
            if isinstance(sila_err, DefinedExecutionError):
                # if defined execution error is not allowed for this call (-> server doesn't follow specs)
                if sila_err.fully_qualified_identifier not in allowed_errors:
                    raise UndefinedExecutionError(f"{sila_err.fully_qualified_identifier}: {sila_err.message}")
                # if client has an error class registered for this error
                if sila_err.fully_qualified_identifier in client._registered_defined_execution_error_classes:
                    raise client._registered_defined_execution_error_classes[sila_err.fully_qualified_identifier](
                        sila_err.message
                    )
            raise sila_err  # plain defined execution error, undefined execution error, validation or framework error
        raise SilaConnectionError(ex)  # server didn't send a SiLA error -> connection error
